import os
import pkg_resources
from unittest import TestCase
from cyvcf2 import VCF
from vafator.annotator import Annotator
import vafator.tests.utils as test_utils


class TestAnnotator(TestCase):

    def test_annotator(self):
        input_file = pkg_resources.resource_filename(__name__, "resources/test1.vcf")
        output_vcf = pkg_resources.resource_filename(__name__, "resources/results/test_annotator1_output.vcf")
        bam1 = pkg_resources.resource_filename(__name__, "resources/COLO_829_n1.bam")
        bam2 = pkg_resources.resource_filename(__name__, "resources/COLO_829_t1.bam")
        annotator = Annotator(
            input_vcf=input_file, output_vcf=output_vcf, input_bams={"normal": [bam1], "tumor": [bam2]})
        annotator.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output)

        info_annotations = test_utils._get_info_fields(output_vcf)
        self.assertTrue("tumor_af" in info_annotations)
        self.assertTrue("normal_af" in info_annotations)
        self.assertTrue("tumor_ac" in info_annotations)
        self.assertTrue("normal_ac" in info_annotations)
        self.assertTrue("tumor_dp" in info_annotations)
        self.assertTrue("normal_dp" in info_annotations)

    def test_annotator_with_multiple_bams(self):
        input_file = pkg_resources.resource_filename(__name__, "resources/test1.vcf")
        output_vcf = pkg_resources.resource_filename(__name__, "resources/results/test_annotator1_output.vcf")
        bam1 = pkg_resources.resource_filename(__name__, "resources/COLO_829_n1.bam")
        bam2 = pkg_resources.resource_filename(__name__, "resources/COLO_829_t1.bam")
        annotator = Annotator(
            input_vcf=input_file, output_vcf=output_vcf, input_bams={"normal": [bam1, bam2], "tumor": [bam1, bam2]})
        annotator.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output)

        info_annotations = test_utils._get_info_fields(output_vcf)
        self.assertTrue("tumor_af_1" in info_annotations)
        self.assertTrue("normal_af_1" in info_annotations)
        self.assertTrue("tumor_ac_1" in info_annotations)
        self.assertTrue("normal_ac_1" in info_annotations)
        self.assertTrue("tumor_dp_1" in info_annotations)
        self.assertTrue("normal_dp_1" in info_annotations)
        self.assertTrue("tumor_af_2" in info_annotations)
        self.assertTrue("normal_af_2" in info_annotations)
        self.assertTrue("tumor_ac_2" in info_annotations)
        self.assertTrue("normal_ac_2" in info_annotations)
        self.assertTrue("tumor_dp_2" in info_annotations)
        self.assertTrue("normal_dp_2" in info_annotations)

    def test_annotator_with_prefix(self):
        input_file = pkg_resources.resource_filename(__name__, "resources/test1.vcf")
        output_vcf = pkg_resources.resource_filename(__name__, "resources/results/test_annotator1_output.vcf")
        bam1 = pkg_resources.resource_filename(__name__, "resources/COLO_829_n1.bam")
        bam2 = pkg_resources.resource_filename(__name__, "resources/COLO_829_t1.bam")
        annotator = Annotator(
            input_vcf=input_file, output_vcf=output_vcf,
            input_bams={"RNA_normal": [bam1, bam2], "RNA_tumor": [bam1, bam2]})
        annotator.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output)

        info_annotations = test_utils._get_info_fields(output_vcf)
        self.assertTrue("RNA_tumor_af_1" in info_annotations)
        self.assertTrue("RNA_normal_af_1" in info_annotations)
        self.assertTrue("RNA_tumor_ac_1" in info_annotations)
        self.assertTrue("RNA_normal_ac_1" in info_annotations)
        self.assertTrue("RNA_tumor_dp_1" in info_annotations)
        self.assertTrue("RNA_normal_dp_1" in info_annotations)
        self.assertTrue("RNA_tumor_af_2" in info_annotations)
        self.assertTrue("RNA_normal_af_2" in info_annotations)
        self.assertTrue("RNA_tumor_ac_2" in info_annotations)
        self.assertTrue("RNA_normal_ac_2" in info_annotations)
        self.assertTrue("RNA_tumor_dp_2" in info_annotations)
        self.assertTrue("RNA_normal_dp_2" in info_annotations)

    def test_annotator_with_mnvs(self):
        input_file = pkg_resources.resource_filename(__name__, "resources/test_tumor_normal.vcf")
        output_vcf = pkg_resources.resource_filename(__name__, "resources/results/test_tumor_normal_output.vcf")
        bam1 = pkg_resources.resource_filename(__name__, "resources/COLO_829_n1.bam")
        bam2 = pkg_resources.resource_filename(__name__, "resources/COLO_829_t1.bam")
        annotator = Annotator(
            input_vcf=input_file, output_vcf=output_vcf,
            input_bams={"RNA_normal": [bam1, bam2], "RNA_tumor": [bam1, bam2]})
        annotator.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output)

        info_annotations = test_utils._get_info_fields(output_vcf)
        self.assertTrue("RNA_tumor_af_1" in info_annotations)
        self.assertTrue("RNA_normal_af_1" in info_annotations)
        self.assertTrue("RNA_tumor_ac_1" in info_annotations)
        self.assertTrue("RNA_normal_ac_1" in info_annotations)
        self.assertTrue("RNA_tumor_dp_1" in info_annotations)
        self.assertTrue("RNA_normal_dp_1" in info_annotations)
        self.assertTrue("RNA_tumor_af_2" in info_annotations)
        self.assertTrue("RNA_normal_af_2" in info_annotations)
        self.assertTrue("RNA_tumor_ac_2" in info_annotations)
        self.assertTrue("RNA_normal_ac_2" in info_annotations)
        self.assertTrue("RNA_tumor_dp_2" in info_annotations)
        self.assertTrue("RNA_normal_dp_2" in info_annotations)

    def _get_info_at(self, input_file, chromosome, position, annotation):
        vcf = VCF(input_file)
        self.assertIsNotNone(vcf)
        for v in vcf:
            if v.POS == position and v.CHROM == chromosome:
                vcf.close()
                return v.INFO.get(annotation)
        vcf.close()
        return {}

    def test_nist(self):
        input_file = pkg_resources.resource_filename(
            __name__, "resources/project.NIST.hc.snps.indels.chr1_1000000_2000000.vcf")
        output_vcf = pkg_resources.resource_filename(
            __name__, "resources/results/project.NIST.hc.snps.indels.chr1_1000000_2000000.vaf.vcf")
        bam_file = pkg_resources.resource_filename(
            __name__,
            "resources/project.NIST_NIST7035_H7AP8ADXX_TAAGGCGA_1_NA12878.bwa.markDuplicates.chr1_1000000_2000000.bam")
        annotator = Annotator(input_vcf=input_file, output_vcf=output_vcf, input_bams={"normal": [bam_file]})
        annotator.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output)

        info_annotations = test_utils._get_info_fields(output_vcf)
        self.assertTrue("normal_af" in info_annotations)
        self.assertTrue("normal_ac" in info_annotations)
        self.assertTrue("normal_dp" in info_annotations)


from dataclasses import dataclass
from typing import List
from cyvcf2 import Variant
from pysam.libcalignmentfile import IteratorColumnRegion, AlignmentFile


@dataclass
class VafatorVariant:
    chromosome: str
    position: int
    reference: str
    alternative: List[str]

    def is_snp(self):
        return len(self.reference) == 1 and len(self.alternative[0]) == 1

    def is_insertion(self):
        return len(self.reference) == 1 and len(self.alternative[0]) > 1

    def is_deletion(self):
        return len(self.alternative[0]) == 1 and len(self.reference) > 1


def build_variant(variant: Variant) -> VafatorVariant:
    return VafatorVariant(
        chromosome=variant.CHROM,
        position=variant.POS,
        reference=variant.REF,
        alternative=variant.ALT
    )


def get_variant_pileup(
        variant: VafatorVariant, bam: AlignmentFile, min_base_quality, min_mapping_quality) -> IteratorColumnRegion:
    chromosome = variant.chromosome
    position = variant.position
    # this function returns the pileups at all positions covered by reads covered the queried position
    # approximately +- read size bp
    return bam.pileup(contig=chromosome, start=position - 1, stop=position, truncate=True,
                      min_base_quality=min_base_quality,
                      min_mapping_quality=min_mapping_quality)


@dataclass
class CoverageMetrics:
    ac: dict
    dp: int


def get_metrics(variant: VafatorVariant, pileups: IteratorColumnRegion) -> CoverageMetrics:
    if variant.is_snp():
        return get_snv_metrics(variant, pileups)
    elif variant.is_insertion():
        return get_insertion_metrics(variant, pileups)
    elif variant.is_deletion():
        return get_deletion_metrics(variant, pileups)
    return None


def get_insertion_metrics(variant: VafatorVariant, pileups: IteratorColumnRegion) -> CoverageMetrics:
    ac = {alt.upper(): 0 for alt in variant.alternative}
    dp = 0
    position = variant.position
    insertion_length = len(variant.alternative[0]) - len(variant.reference)
    insertion = variant.alternative[0][1:]
    try:
        pileup = next(pileups)
        for r in pileup.pileups:
            dp += 1
            if r.indel > 0:
                # read with an insertion
                start = r.alignment.reference_start
                relative_position = 0
                for cigar_type, cigar_length in r.alignment.cigartuples:
                    if cigar_type in [0, 2, 3, 7, 8]:  # consumes reference M, D, N, =, X
                        start += cigar_length
                        if start > position:
                            break
                    if cigar_type in [0, 1, 4, 7, 8]:  # consumes query M, I, S, =, X
                        relative_position += cigar_length
                    if cigar_type == 1:  # does not count I
                        insertion_in_query = r.alignment.query[relative_position:relative_position + insertion_length]
                        if start == position and cigar_length == insertion_length and \
                                insertion == insertion_in_query:
                            ac[variant.alternative[0].upper()] = ac[variant.alternative[0].upper()] + 1
    except StopIteration:
        # no reads
        pass
    return CoverageMetrics(ac=ac, dp=dp)


def get_deletion_metrics(variant: VafatorVariant, pileups: IteratorColumnRegion) -> CoverageMetrics:
    ac = {alt.upper(): 0 for alt in variant.alternative}
    dp = 0
    position = variant.position
    deletion_length = len(variant.reference) - len(variant.alternative[0])
    try:
        pileup = next(pileups)
        for r in pileup.pileups:
            dp += 1
            if r.indel < 0:
                # read with an deletion
                start = r.alignment.reference_start
                for cigar_type, cigar_length in r.alignment.cigartuples:
                    if cigar_type in [0, 3, 7, 8]:  # consumes reference M, N, =, X
                        start += cigar_length
                    elif cigar_type == 2:  # D
                        if start == position and cigar_length == deletion_length:
                            ac[variant.alternative[0].upper()] = ac[variant.alternative[0].upper()] + 1
                        else:
                            start += cigar_length
                    if start > position:
                        break
    except StopIteration:
        # no reads
        pass
    return CoverageMetrics(ac=ac, dp=dp)


def _initialize_empty_count(bases_counts, base):
    if bases_counts.get(base) is None:
        bases_counts[base] = 0
    return bases_counts


def get_snv_metrics(variant: VafatorVariant, pileups: IteratorColumnRegion) -> CoverageMetrics:
    dp = 0
    ac = {alt.upper(): 0 for alt in variant.alternative}
    ac[variant.reference.upper()] = 0
    try:
        pileup = next(pileups)
        for s in pileup.get_query_sequences():
            dp += 1
            base_upper_case = s.upper()
            if base_upper_case in ac:
                ac[base_upper_case] = ac[base_upper_case] + 1
    except StopIteration:
        # no reads
        pass
    return CoverageMetrics(ac=ac, dp=dp)



from cyvcf2 import VCF, Writer
import vafator
import datetime
import json
import random


class MultiallelicFilter(object):

    vafator_header = {
        "name": "multiallelic-filter",
        "version": vafator.VERSION,
        "date": datetime.datetime.now().ctime(),
        "timestamp": datetime.datetime.now().timestamp(),
    }

    multiallelic_annotation_name = "multiallelic"

    def __init__(self, input_vcf, output_vcf, tumor_sample_name='tumor'):
        """
        :param input_vcf: the input VCF file
        :param output_vcf: the file path of the output VCF
        """
        self.tumor_sample_name = tumor_sample_name
        self.vcf = VCF(input_vcf)
        # sets a line in the header with the command used to annotate the file
        self.vafator_header["input_vcf"] = input_vcf
        self.vafator_header["output_vcf"] = output_vcf
        self.vcf.add_to_header("##vafator_command_line={}".format(json.dumps(self.vafator_header)))
        # adds to the header all the names of the annotations
        self.vcf.add_info_to_header({'ID': self.multiallelic_annotation_name, 'Type': 'String', 'Number': '.',
                 'Description': "Indicates multiallelic variants filtered and their frequencies if any (e.g.: T,0.12)"})
        self.vcf_writer = Writer(output_vcf, self.vcf)

    def run(self):
        batch = []
        prev_variant = None
        for variant in self.vcf:
            if prev_variant is None:
                # first time stores the variant and moves on
                prev_variant = variant
                continue
            # considers only SNVs with same chromosome, position and reference
            if variant.is_snp and variant.CHROM == prev_variant.CHROM and variant.POS == prev_variant.POS and \
                    variant.REF == prev_variant.REF:
                af1 = self.get_tumor_af(prev_variant)
                af2 = self.get_tumor_af(variant)
                # keeps the variant with the highest AF
                if af1 < af2:
                    self.set_multiallelic_annotation(variant, prev_variant.ALT[0], af1)
                    prev_variant = variant
                elif af1 > af2:
                    self.set_multiallelic_annotation(prev_variant, variant.ALT[0], af2)
                elif af1 == af2:
                    # chooses a variant at random
                    alt1 = variant.ALT[0]
                    alt2 = prev_variant.ALT[0]
                    prev_variant = random.sample([variant, prev_variant], k=1)[0]
                    self.set_multiallelic_annotation(prev_variant, alt1 if alt1 != prev_variant.ALT[0] else alt2, af1)
                continue

            # write previous variant and stores current for next iteration
            batch.append(prev_variant)
            prev_variant = variant

            if len(batch) >= 1000:
                self._write_batch(batch)
                batch = []
        # do not forget to add the last variant!
        batch.append(prev_variant)
        self._write_batch(batch)

        self.vcf_writer.close()
        self.vcf.close()

    def set_multiallelic_annotation(self, variant, alt, af):
        variant.INFO[self.multiallelic_annotation_name] = \
            ",".join(variant.INFO.get(self.multiallelic_annotation_name, "").split(",") + [alt, str(af)])

    def get_tumor_af(self, prev_variant):
        return prev_variant.INFO.get("{}_af".format(self.tumor_sample_name), 0.0)

    def _write_batch(self, batch):
        for v in batch:
            self.vcf_writer.write_record(v)



from cyvcf2 import VCF


def _get_count_variants(input_file):
    vcf = VCF(input_file)
    n_variants = 0
    for v in vcf:
        n_variants += 1
    vcf.close()
    return n_variants


def _get_info_fields(input_file):
    vcf = VCF(input_file)
    return [h.info().get("ID") for h in vcf.header_iter() if h['HeaderType'] == 'INFO']



from unittest import TestCase
import pkg_resources
import pysam
from vafator.pileups import get_variant_pileup, VafatorVariant, get_metrics


class TestPileups(TestCase):

    min_base_quality = 0
    min_mapping_quality = 0
    bam_file = pkg_resources.resource_filename(
        __name__,
        "resources/project.NIST_NIST7035_H7AP8ADXX_TAAGGCGA_1_NA12878.bwa.markDuplicates.chr1_1000000_2000000.bam")
    bam_reader = pysam.AlignmentFile(bam_file)

    def test_snv_metrics(self):
        variant = VafatorVariant(chromosome="chr1", position=1017341, reference="G", alternative=["T"])
        self._assert_metrics(variant=variant, expected_ac={'G': 5, 'T': 6}, expected_dp=11)
        self._assert_metrics(variant=variant, expected_ac={'G': 1, 'T': 2}, expected_dp=3, min_base_quality=40)

    def test_snv_metrics_2(self):
        variant = VafatorVariant(chromosome="chr1", position=1018144, reference="T", alternative=["C"])
        self._assert_metrics(variant=variant, expected_ac={'C': 9, 'T': 11}, expected_dp=20)
        self._assert_metrics(variant=variant, expected_ac={'C': 3, 'T': 4}, expected_dp=7, min_base_quality=40)
        self._assert_metrics(variant=variant, expected_ac={'C': 0, 'T': 0}, expected_dp=0, min_mapping_quality=65)

    def test_insertion_metrics(self):
        # variant called in the VCF shows no read support (!?)
        variant = VafatorVariant(chromosome="chr1", position=1247578, reference="T", alternative=["TGG"])
        self._assert_metrics(variant=variant, expected_ac={'TGG': 0}, expected_dp=3)
        # there is one read supporting this insertion of 3 Gs
        variant = VafatorVariant(chromosome="chr1", position=1247578, reference="T", alternative=["TGGG"])
        self._assert_metrics(variant=variant, expected_ac={'TGGG': 1}, expected_dp=3)
        # this ensures that the insertion sequence is checked not only the insertion length!
        variant = VafatorVariant(chromosome="chr1", position=1247578, reference="T", alternative=["TGGA"])
        self._assert_metrics(variant=variant, expected_ac={'TGGA': 0}, expected_dp=3)
        # there is one read supporting this insertion of 4 Gs
        variant = VafatorVariant(chromosome="chr1", position=1247578, reference="T", alternative=["TGGGG"])
        self._assert_metrics(variant=variant, expected_ac={'TGGGG': 1}, expected_dp=3)
        # there is no read supporting an insertion of 5 Gs
        variant = VafatorVariant(chromosome="chr1", position=1247578, reference="T", alternative=["TGGGGG"])
        self._assert_metrics(variant=variant, expected_ac={'TGGGGG': 0}, expected_dp=3)

    def test_insertion_metrics_2(self):
        variant = VafatorVariant(chromosome="chr1", position=1594199, reference="C", alternative=["CT"])
        self._assert_metrics(variant=variant, expected_ac={'CT': 9}, expected_dp=11)
        self._assert_metrics(variant=variant, expected_ac={'CT': 5}, expected_dp=7, min_mapping_quality=40)
        self._assert_metrics(variant=variant, expected_ac={'CT': 4}, expected_dp=4, min_base_quality=40)

    def test_deletion_metrics(self):
        variant = VafatorVariant(chromosome="chr1", position=1510035, reference="GGC", alternative=["G"])
        self._assert_metrics(variant=variant, expected_ac={'G': 12}, expected_dp=13)
        self._assert_metrics(variant=variant, expected_ac={'G': 0}, expected_dp=0, min_mapping_quality=61)
        self._assert_metrics(variant=variant, expected_ac={'G': 3}, expected_dp=3, min_base_quality=40)
        # deletions with a reference sequence not matching the reference would be matched
        # vafator expects correct indel calls
        variant = VafatorVariant(chromosome="chr1", position=1510035, reference="GCC", alternative=["G"])
        self._assert_metrics(variant=variant, expected_ac={'G': 12}, expected_dp=13)

    def _assert_metrics(self, variant: VafatorVariant, expected_ac, expected_dp,
                        min_base_quality=0, min_mapping_quality=0):
        pileups = get_variant_pileup(
            variant=variant, bam=self.bam_reader,
            min_base_quality=min_base_quality, min_mapping_quality=min_mapping_quality)
        coverage_metrics = get_metrics(variant=variant, pileups=pileups)
        self.assertEqual(expected_ac, coverage_metrics.ac)
        self.assertEqual(expected_dp, coverage_metrics.dp)



#!/usr/bin/env python
import argparse
import sys
import logging
import vafator
from vafator.annotator import Annotator
from vafator.multiallelic_filter import MultiallelicFilter
from vafator.vafator2decifer import run_vafator2decifer

epilog = "Copyright (c) 2019-2021 TRON gGmbH (See LICENSE for licensing details)"


def annotator():

    # set up logger
    parser = argparse.ArgumentParser(description="vafator v{}".format(vafator.VERSION),
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     epilog=epilog)
    parser.add_argument("--input-vcf", dest="input_vcf", action="store", help="The VCF to annotate", required=True)
    parser.add_argument("--output-vcf", dest="output_vcf", action="store", help="The annotated VCF", required=True)
    parser.add_argument('--bam', action='append', nargs=2,
                        metavar=('sample_name', 'bam_file'), default=[],
                        help='A sample name and a BAM file. Can be used multiple times to input multiple samples and '
                             'multiple BAM files. The same sample name can be used multiple times with different BAMs, '
                             'this will treated as replicates.')
    parser.add_argument("--normal-bams", dest="normal_bams", nargs='+', action="store", default=[],
                        help="Whitespace-separated list of normal BAMs to analyse")
    parser.add_argument("--tumor-bams", dest="tumor_bams", nargs='+', action="store", default=[],
                        help="Whitespace-separated list of tumor BAMs to analyse")
    parser.add_argument("--mapping-quality", dest="mapping_quality", action="store", type=int, default=1,
                        help="All reads with a mapping quality below this threshold will be filtered out")
    parser.add_argument("--base-call-quality", dest="base_call_quality", action="store", type=int, default=30,
                        help="All bases with a base call quality below this threshold will be filtered out")

    args = parser.parse_args()

    logging.info("Vafator starting...")

    bams = {}
    for sample_name, bam in args.bam:
        if sample_name in bams:
            bams[sample_name].append(bam)
        else:
            bams[sample_name] = [bam]
    for bam in args.tumor_bams:
        if "tumor" in bams:
            bams["tumor"].append(bam)
        else:
            bams["tumor"] = [bam]
    for bam in args.normal_bams:
        if "normal" in bams:
            bams["normal"].append(bam)
        else:
            bams["normal"] = [bam]

    if len(bams) == 0:
        raise ValueError("Please, provide at least one bam file through either --bam, --tumor-bams or --normal-bams")

    try:
        annotator = Annotator(
            input_vcf=args.input_vcf,
            output_vcf=args.output_vcf,
            input_bams=bams,
            mapping_qual_thr=args.mapping_quality,
            base_call_qual_thr=args.base_call_quality
        )
        annotator.run()
    except Exception as e:
        logging.error(str(e))
        sys.exit(-1)
    logging.info("Vafator finished!")


def multiallelics_filter():

    # set up logger
    parser = argparse.ArgumentParser(description="vafator v{}".format(vafator.VERSION),
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter, epilog=epilog)
    parser.add_argument("--input-vcf", dest="input_vcf", action="store", help="The VCF to annotate", required=True)
    parser.add_argument("--output-vcf", dest="output_vcf", action="store", help="The annotated VCF", required=True)
    parser.add_argument("--tumor-sample-name", dest="tumor_sample_name", action="store",
                        help='The tumor sample name (will look for annotation ${SAMPLE_NAME}_af)', default='tumor')
    args = parser.parse_args()

    logging.info("Vafator multiallelic filter starting...")
    try:
        filter = MultiallelicFilter(
            input_vcf=args.input_vcf,
            output_vcf=args.output_vcf,
            tumor_sample_name=args.tumor_sample_name
        )
        filter.run()
    except Exception as e:
        logging.error(str(e))
        sys.exit(-1)
    logging.info("Vafator multiallelic filter finished!")


def vafator2decifer():
    parser = argparse.ArgumentParser(description='Generate input for Decifer using VCF file and HATCHet CNA file')
    parser.add_argument("-V", "--vcf_file", required=True, type=str, help="single or multi-sample VCF file")
    parser.add_argument("-S", "--samples", required=True, type=str,
                        help="comma separated list of sample name prefixes to use for VAFator annotations, "
                             "eg: primary_tumor,metastasis_tumor; the annotations primary_tumor_ac, primary_tumor_dp, "
                             "etc. will be expected to exist")
    parser.add_argument("-C", "--cna_file", required=True, type=str, help="HATCHet CNA file: best.seg.ucn ")
    parser.add_argument("-O", "--out_dir", required=True, default="./", type=str,
                        help="directory for printing files; please make unique for each patient!")
    parser.add_argument("-M", "--min_depth", required=True, type=int, help="minimum depth PER sample")
    parser.add_argument("-A", "--min_alt_depth", required=True, type=int,
                        help="minimum depth of ALT allele in at least one sample")
    parser.add_argument("-N", "--max_CN", required=False, default=6, type=int,
                        help="maximum total copy number for each observed clone")
    args = parser.parse_args()
    run_vafator2decifer(args)



import os
import pkg_resources
from unittest import TestCase
from cyvcf2 import VCF
from vafator.multiallelic_filter import MultiallelicFilter
import vafator.tests.utils as test_utils


class TestMultiallelicFilter(TestCase):

    def setUp(self):
        # self.reference_genome_reader = TestReferenceGenomeReader()
        pass

    def test_no_variants_filtered(self):
        input_file = pkg_resources.resource_filename(__name__, "resources/test1.vcf")
        output_vcf = pkg_resources.resource_filename(__name__, "resources/results/test1_output.vcf")
        multiallelic_filter = MultiallelicFilter(input_vcf=input_file, output_vcf=output_vcf, tumor_sample_name='tumor')
        multiallelic_filter.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output)

    def test_two_variants_filtered(self):
        input_file = pkg_resources.resource_filename(__name__, "resources/test2.vcf")
        output_vcf = pkg_resources.resource_filename(__name__, "resources/results/test2_output.vcf")
        multiallelic_filter = MultiallelicFilter(input_vcf=input_file, output_vcf=output_vcf, tumor_sample_name='tumor')
        multiallelic_filter.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output + 3, "input:{}; output:{}".format(
            n_variants_input, n_variants_output))

        af1 = self._get_info_at(output_vcf, chromosome="chr4", position=1235, annotation='tumor_af')
        self.assertTrue(af1, 0.2)
        multiallelic1 = self._get_info_at(output_vcf, chromosome="chr4", position=1235, annotation='multiallelic')
        self.assertTrue(multiallelic1, "T,0.1")

        af2 = self._get_info_at(output_vcf, chromosome="chr6", position=1235, annotation='tumor_af')
        self.assertTrue(af2, 0.2)
        multiallelic2 = self._get_info_at(output_vcf, chromosome="chr6", position=1235, annotation='multiallelic')
        self.assertTrue(multiallelic2, "A,0.01")

        af3 = self._get_info_at(output_vcf, chromosome="chr6", position=1234, annotation='tumor_af')
        self.assertTrue(af3, 0.5)

    def test_different_reference_is_kept(self):
        input_file = pkg_resources.resource_filename(__name__, "resources/test3.vcf")
        output_vcf = pkg_resources.resource_filename(__name__, "resources/results/test3_output.vcf")
        multiallelic_filter = MultiallelicFilter(input_vcf=input_file, output_vcf=output_vcf, tumor_sample_name='tumor')
        multiallelic_filter.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output)
        self.assertTrue(n_variants_input == 2)

    def test_three_multiallelics(self):
        input_file = pkg_resources.resource_filename(__name__, "resources/test4.vcf")
        output_vcf = pkg_resources.resource_filename(__name__, "resources/results/test4_output.vcf")
        multiallelic_filter = MultiallelicFilter(input_vcf=input_file, output_vcf=output_vcf, tumor_sample_name='tumor')
        multiallelic_filter.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output + 2)
        self.assertTrue(n_variants_output == 1)

    def test_equal_af(self):
        input_file = pkg_resources.resource_filename(__name__, "resources/test5.vcf")
        output_vcf = pkg_resources.resource_filename(__name__, "resources/results/test5_output.vcf")
        multiallelic_filter = MultiallelicFilter(input_vcf=input_file, output_vcf=output_vcf, tumor_sample_name='tumor')
        multiallelic_filter.run()

        self.assertTrue(os.path.exists(output_vcf))
        n_variants_input = test_utils._get_count_variants(input_file)
        n_variants_output = test_utils._get_count_variants(output_vcf)
        self.assertTrue(n_variants_input == n_variants_output + 2)
        self.assertTrue(n_variants_output == 1)

    def _get_info_at(self, input_file, chromosome, position, annotation):
        vcf = VCF(input_file)
        self.assertIsNotNone(vcf)
        for v in vcf:
            if v.POS == position and v.CHROM == chromosome:
                vcf.close()
                return v.INFO.get(annotation)
        vcf.close()
        return {}


from setuptools import find_packages, setup
import vafator


VERSION = vafator.VERSION


# parses requirements from file
with open("requirements.txt") as f:
    required = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

# Build the Python package
setup(
    name='vafator',
    version=VERSION,
    packages=find_packages(exclude=["legacy"]),
    entry_points={
        'console_scripts': [
            'vafator=vafator.command_line:annotator',
            'multiallelics-filter=vafator.command_line:multiallelics_filter',
            'vafator2decifer=vafator.command_line:vafator2decifer'
        ],
    },
    author="TRON - Translational Oncology at the University Medical Center of the Johannes Gutenberg University Mainz"
    "- Computational Medicine group",
    author_email='pablo.riesgoferreiro@tron-mainz.de',
    description='Annotate a VCF file with AF, AD and DP from tumor and normal BAMs',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tron-bioinformatics/vafator",
    requires=[],
    install_requires=required,
    classifiers=[
        'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix"
      ],
    python_requires='>=3.7',
    license='MIT'
)


import pysam
from cyvcf2 import VCF, Writer, Variant
import os
import vafator
import datetime
import json
from vafator.pileups import get_variant_pileup, build_variant, get_metrics


class Annotator(object):

    vafator_header = {
        "name": "vafator",
        "version": vafator.VERSION,
        "date": datetime.datetime.now().ctime(),
        "timestamp": datetime.datetime.now().timestamp(),
    }

    def __init__(self, input_vcf: str, output_vcf: str, input_bams: dict, mapping_qual_thr=0, base_call_qual_thr=29):

        self.mapping_quality_threshold = mapping_qual_thr
        self.base_call_quality_threshold = base_call_qual_thr
        self.vcf = VCF(input_vcf)
        # sets a line in the header with the command used to annotate the file
        self.vafator_header["input_vcf"] = input_vcf
        self.vafator_header["output_vcf"] = output_vcf
        self.vafator_header["bams"] = ";".join(["{}:{}".format(s, ",".join(b)) for s, b in input_bams.items()])
        self.vafator_header["mapping_quality_threshold"] = mapping_qual_thr
        self.vafator_header["base_call_quality_threshold"] = base_call_qual_thr
        self.vcf.add_to_header("##vafator_command_line={}".format(json.dumps(self.vafator_header)))
        # adds to the header all the names of the annotations
        for a in Annotator._get_headers(input_bams):
            self.vcf.add_info_to_header(a)
        self.vcf_writer = Writer(output_vcf, self.vcf)
        self.bam_readers = {s : [pysam.AlignmentFile(b, "rb") for b in bams] for s, bams in input_bams.items()}

    @staticmethod
    def _get_headers(input_bams: dict):
        headers = []

        for s, bams in input_bams.items():
            headers.append({
                'ID': "{}_af".format(s),
                'Description': "Allele frequency for the alternate alleles in the {} sample/s".format(s),
                'Type': 'Float',
                'Number': 'A'
            })
            headers.append({
                'ID': "{}_dp".format(s),
                'Description': "Total depth of coverage in the {} sample/s (independent of alleles)".format(s),
                'Type': 'Float',
                'Number': '1'
            })
            headers.append({
                'ID': "{}_ac".format(s),
                'Description': "Allele count for the alternate alleles in the {} sample/s".format(s),
                'Type': 'Integer',
                'Number': 'A'
            })

            if len(bams) > 1:
                for i, bam in enumerate(bams, start=1):
                    n = os.path.basename(bam).split(".")[0]
                    headers = headers + [
                        {'ID': "{}_af_{}".format(s, i),
                         'Description': "Allele frequency for the alternate alleles in the {} sample {}".format(s, n),
                         'Type': 'Float', 'Number': 'A'},
                        {'ID': "{}_dp_{}".format(s, i),
                         'Description': "Depth of coverage in the {} sample {} (independent of alleles)".format(s, n),
                         'Type': 'Float', 'Number': '1'},
                        {'ID': "{}_ac_{}".format(s, i),
                         'Description': "Allele count for the alternate alleles in the {} sample {}".format(s, n),
                         'Type': 'Integer', 'Number': 'A'}
                    ]
        return headers

    def _write_batch(self, batch):
        for v in batch:
            self.vcf_writer.write_record(v)

    def _add_stats(self, variant: Variant):

        global_dp = 0
        global_ac = {}
        vafator_variant = build_variant(variant)
        for sample, bams in self.bam_readers.items():
            for i, bam in enumerate(bams):
                pileups = get_variant_pileup(
                    variant=vafator_variant, bam=bam,
                    min_base_quality=self.base_call_quality_threshold,
                    min_mapping_quality=self.mapping_quality_threshold)
                coverage_metrics = get_metrics(variant=vafator_variant, pileups=pileups)
                if coverage_metrics is not None:
                    if len(bams) > 1:
                        variant.INFO["{}_af_{}".format(sample, i + 1)] = ",".join(
                            [str(self._calculate_af(coverage_metrics.ac[alt], coverage_metrics.dp)) for alt in variant.ALT])
                        variant.INFO["{}_ac_{}".format(sample, i + 1)] = ",".join([str(coverage_metrics.ac[alt]) for alt in variant.ALT])
                        variant.INFO["{}_dp_{}".format(sample, i + 1)] = coverage_metrics.dp
                    for alt in variant.ALT:
                        global_ac[alt] = global_ac.get(alt, 0) + coverage_metrics.ac[alt]
                    global_dp += coverage_metrics.dp

            variant.INFO["{}_af".format(sample)] = ",".join([str(self._calculate_af(global_ac.get(alt, 0), global_dp)) for alt in variant.ALT])
            variant.INFO["{}_ac".format(sample)] = ",".join([str(global_ac.get(alt, 0)) for alt in variant.ALT])
            variant.INFO["{}_dp".format(sample)] = global_dp

    def _calculate_af(self, ac, dp):
        return float(ac) / dp if dp > 0 else 0.0

    def run(self):
        batch = []
        variant: Variant
        for variant in self.vcf:
            # gets the counts of all bases across all BAMs
            self._add_stats(variant)

            batch.append(variant)
            if len(batch) >= 1000:
                self._write_batch(batch)
                batch = []
        if len(batch) > 0:
            self._write_batch(batch)

        self.vcf_writer.close()
        self.vcf.close()
        for _, bams in self.bam_readers.items():
            for bam in bams:
                bam.close()



#!/usr/bin/python -tt

"""
Please type "python3 vcf_2_decifer.py --help" for a the full list of options

This is a python script that takes as input (1) a multi-sample VCF, e.g. variants called in multiple tumor samples from a single patient, (2) a file containing copy number aberration information, e.g. the "best.seg.ucn" file that is output by the program HATCHet. The sample names must agree in both files!

This script also uses cyvcf2 to parse VCF files efficiently, and the python implementation of bedtools to find which CNA interval each SNV overlaps with. Please make a conda environment containing these python modules before running this script.
For instance, do the following to run this script:

conda create -n vcf_bedtools pybedtools cyvcf2 pandas -y
conda activate vcf_bedtools
python vcf_2_decifer.py [OPTIONS]
"""

import re
import sys
import pybedtools as pbt
from cyvcf2 import VCF, Variant
import os
import pandas as pd
import numpy as np
from collections import defaultdict
import argparse


def filterByDepth(variant: Variant, Filter, samples):
    PASS = 1
    missing = 0
    for s in samples:
        # filter if genotype has low depth or is missing
        if variant.INFO["{}_dp".format(s)] < Filter['MinDepth']:
            missing += 1
    # filter if alt allele isn't greater than the specified threshold in at least one sample
    if not any(np.greater_equal([variant.INFO["{}_ac".format(s)] for s in samples], Filter['MinDepthAltAllele'])):
        missing += 1
    # (gt_alt_depths[i] < Filter['MinDepthAltAllele'])
    if missing > 0:
        PASS = 0
    return (PASS)


def compute_ref_var_depths(vcf, FilterDP, samples):
    ref_var_depths = defaultdict(list)
    # ref_var_depths[char_label] = list of (ref,alt) tuples, one for each sample, in same order as vcf.samples
    variant: Variant
    for variant in vcf:
        if len(variant.ALT) == 1 and variant.var_type == "snp":
            PASS = filterByDepth(variant, FilterDP, samples)
            # print(np.greater_equal(variant.gt_alt_depths,FilterDP['MinDepthAltAllele']))
            if PASS:
                char_label = ".".join(map(str, [variant.CHROM, variant.POS, variant.REF, variant.ALT[0]]))
                for s in samples:
                    alt = variant.INFO["{}_ac".format(s)]
                    ref = variant.INFO["{}_dp".format(s)] - alt
                    ref_var_depths[char_label].append((ref, alt))
    return ref_var_depths


def print_output(vcf, ref_var_depths, cna_overlaps, outdir):
    char_index = 0
    chars = ref_var_depths.keys() & cna_overlaps.keys()
    header = [str(len(chars)) + " #characters"]
    header.append(str(len(vcf.samples)) + " #samples")
    header.append("#sample_index\tsample_label\tcharacter_index\tcharacter_label\tref\tvar")
    print(header)
    with open(f"{outdir}/decifer.input.tsv", 'w') as out:
        print("\n".join(header), file=out)
        for char_label in ref_var_depths:
            if char_label in cna_overlaps:
                for i in range(len(vcf.samples)):
                    r, v = ref_var_depths[char_label][i][0], ref_var_depths[char_label][i][1]
                    to_print = [i, vcf.samples[i], char_index, char_label, r, v]
                    cnas = cna_overlaps[char_label][i]
                    to_print.extend(cnas)
                    print("\t".join(map(str, to_print)), file=out)
                    # print(i, vcf.samples[i], char_index, char_label, r, v)
                char_index += 1


def print_purities(cna_df, sample_index, num_samples, outdir):
    purities = {}
    for i, row in cna_df.head(num_samples + 1).iterrows():
        purities[row['SAMPLE']] = 1.0 - row['u_normal']
    with open(f"{outdir}/decifer.purity.tsv", 'w') as out:
        for sample in sample_index:
            print(sample_index[sample], purities[sample], file=out, sep="\t")


def filter_high_CN_sites(cn_states_persite, max_CN):
    # returns 0 if site has CN state greater than max_CN
    # cn_states_persite is a list of str "A|B", CN states of maternal/paternal chromosomes
    for i in cn_states_persite:
        j = i.split("|")
        if int(j[0]) + int(j[1]) > max_CN:
            return 0
    return 1


def print_unique_CN_states(cn_states, max_CN, outdir):
    # print unique copy number states for sites that are below the max_CN threshold
    cn_states = tuple(set(cn_states))
    with open(f"{outdir}/cn_states.txt", 'w') as out:
        for value in cn_states:
            PASS = 1
            for i in value:
                if int(i[0]) + int(i[1]) > max_CN:
                    PASS = 0
            if PASS:
                out.write(';'.join([','.join(i) for i in value]) + '\n')


def print_filtered_sites(filtered_sites, cna_overlaps, outdir):
    with open(f"{outdir}/filtered_sites.txt", 'w') as out:
        out.write("\n".join(filtered_sites))
        print(file=out)
    with open(f"{outdir}/filtered_stats.txt", 'w') as out:
        filtered = len(filtered_sites)
        total = len(cna_overlaps.keys())
        print("# sites that were filtered due to copy-number states > max_CN", file=out)
        print("filtered: ", filtered, file=out)
        print("fraction: ", float(filtered / total), file=out)


def overlap_cna_snp(vcf_samples, max_CN, out_dir):
    cna_overlaps = defaultdict(list)
    cn_states_allsites = []  # a list of tuples
    filtered_sites = set()  # sites filtered out because of high CN
    snps = pbt.BedTool(f"{out_dir}/snps.bed")
    # for each sample in VCF, intersect it's SNVs with sample-specific CNAs
    for sample in vcf_samples:
        sample_cnas = pbt.BedTool(f"{out_dir}/{sample}_cna.bed")
        # snps.intersect(sample_cnas, wo=True).saveas(f"snps_cnas_overlap_{sample}.bed")
        bed = snps.intersect(sample_cnas, wo=True)
        for line in bed:
            line = str(line).split()
            # first 5 columns are SNP info (chr,pos_start,pos_end,REF,ALT), rest are CNA info
            # CNA info starts with CHR, START, END, which you don't want, so start at index 8
            # exclude the last index, since bedtools adds this, the number of bp of overlap
            char_label = ".".join([line[0], line[2], line[3], line[4]])
            cns = line[8:-1:2]  # copy-number states, string[start:end:step]
            props = line[9:-1:2]  # copy-number proportions
            if len(cns) != len(props):
                sys.exit("CNA file not formatted correctly!")

            cn_info = defaultdict(float)  # dict with cn_info["A|B"] = proportion
            for c, p in zip(cns, props):
                cn_info[c] += float(p)  # this collapses nonunique CN states

            # returns 0 if CN state too high
            if filter_high_CN_sites(cn_info.keys(), max_CN):
                # store results, converting from dict to a list for later printing
                cna_info = []
                [cna_info.extend([c.split("|")[0], c.split("|")[1], cn_info[c]]) for c in cn_info]
                cna_overlaps[char_label].append(cna_info)
            else:
                filtered_sites.add(char_label)
            # get collection of unique CN states for this SNV site
            tuple_states = [(c.split("|")[0], c.split("|")[1]) for c in cn_info]
            cn_states_allsites.append(tuple(set(tuple_states)))

    return cna_overlaps, cn_states_allsites, filtered_sites


def run_vafator2decifer(args):

    vcf_name = os.path.basename(args.vcf_file)
    vcf = VCF(args.vcf_file, gts012=True)

    # Filtering criteria
    FilterDP = {}
    FilterDP['MinDepth'] = args.min_depth
    FilterDP['MinDepthAltAllele'] = args.min_alt_depth

    samples = args.samples.split(",")
    num_samples = len(samples)
    sample_index = {samples[i]: i for i in range(len(samples))}

    # ref_var_depths[char_label] = list of (ref,alt) tuples, one for each sample, in same order as vcf.samples
    ref_var_depths = compute_ref_var_depths(vcf, FilterDP, samples)

    # print BED file for SNPs
    with open(f"{args.out_dir}/snps.bed", 'w') as out:
        print("chrom\tstart\tend\tREF\tALT", file=out)
        for chr_label in ref_var_depths:
            pos = chr_label.split(".")
            # subtract 1 from position to create interval in BED format
            print(pos[0], int(pos[1]) - 1, int(pos[1]), pos[2], pos[3], sep="\t", file=out)

    # Load in CNA information
    cna_df = pd.read_csv(args.cna_file, sep='\t', index_col=False)
    # print purity information
    print_purities(cna_df, sample_index, num_samples, args.out_dir)

    # prepare BED files for CNA intervals for each sample, for overlapping with SNPs
    for sample in samples:
        df = cna_df[cna_df['SAMPLE'] == sample]
        # consider subtracting 1 from start of interval to be compatible with BED format, leave end interval alone
        df.loc[:, 'START'] = df['START']
        df = df.drop('SAMPLE', axis=1)
        df.to_csv(f"{args.out_dir}/{sample}_cna.bed", index=False, sep="\t")

    # overlap SNPs with CNA intervals for each sample
    # cna_overlaps[char_label] = list of tuples of CNA info (one tuple for each sample, in same order as vcf.samples)
    # this function also prints the observed CN state trees for the generatestatetrees function
    cna_overlaps, cn_states_allsites, filtered_sites = overlap_cna_snp(samples, args.max_CN, args.out_dir)

    # sites may have unique CN states that are duplicate; set them to find unique CN states across sites
    print_unique_CN_states(cn_states_allsites, args.max_CN, args.out_dir)
    print_filtered_sites(filtered_sites, cna_overlaps, args.out_dir)

    print_output(vcf, ref_var_depths, cna_overlaps, args.out_dir)

    os.system(f"rm {args.out_dir}/snps.bed")
    for sample in samples:
        os.system(f"rm {args.out_dir}/{sample}_cna.bed")
